package ch.hepia.exos;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;
import org.junit.Test;

import java.util.Set;
import java.util.function.Function;
import java.util.HashSet;

public class MySetTest {
    @Test
    public void testOf() {
        assertEquals(Set.of(1.0, 2.0, 3.0, 4.0), MySet.of(1.0, 2.0, 3.0, 4.0).toSet());
        assertEquals(Set.of(1.0, 4.0), MySet.of(1.0, 4.0, 1.0, 4.0).toSet());
    }

    @Test
    public void testRange() {
        assertEquals(Set.of(0, 1, 2, 3, 4), MySet.range(0, 5).toSet());
        assertEquals(Set.of(2, 3, 4, 5), MySet.range(2, 6).toSet());
        assertEquals(new HashSet<>(), MySet.range(6, 2).toSet());
    }

    @Test
    public void testMySet() {
        Set<Integer> tmp = Set.of(0, 1, 2, 3);
        assertEquals(tmp, new MySet<Integer>(tmp).toSet());
    }

    @Test
    public void testAdd() {
        MySet<Integer> tmp = new MySet<>();
        tmp.add(5);
        tmp.add(3);
        tmp.add(6);
        assertEquals(Set.of(3, 6, 5), tmp.toSet());
    }

    @Test
    public void testAddAll() {
        MySet<Integer> tmpA = new MySet<>();
        MySet<Integer> tmpB = MySet.of(0, 1, 2, 3, 4, 5);
        tmpA.addAll(tmpB);
        assertEquals(tmpB.toSet(), tmpA.toSet());
    }

    @Test
    public void testClear() {
        MySet<Integer> tmp = MySet.range(0, 6);
        tmp.clear();
        assertEquals(new MySet<>().toSet(), tmp.toSet());
    }

    // there is no test for the toSet method, it's used by all the other tests.

    @Test
    public void testForEach() {
        MySet<Integer> tmpA = new MySet<>();
        MySet<Integer> tmpB = MySet.range(0, 6);
        tmpB.forEach(tmpA::add);
        assertEquals(tmpB.toSet(), tmpA.toSet());
    }

    @Test
    public void testMap() {
        assertEquals(Set.of(1, 2, 3, 4, 5), MySet.range(0, 5).map(i -> i + 1).toSet());
        assertEquals(Set.of(true, false), MySet.range(0, 6).map(i -> i % 2 == 0).toSet());
    }

    @Test
    public void testReduce() {
        assertSame(15, MySet.range(0, 6).reduce(0, (x, y) -> x + y));
        assertSame(120, MySet.range(1, 6).reduce(1, (x, y) -> x * y));
    }

    @Test
    public void testMapReduce() {
        MySet<Integer> tmpA = MySet.range(0, 5);
        MySet<Integer> tmpB = MySet.range(5, 10);
        MySet<Integer> tmpC = MySet.range(10, 15);
        MySet<MySet<Integer>> tmpD = MySet.of(tmpA, tmpB, tmpC);
        assertEquals(MySet.range(0, 15).toSet(),
                tmpD.mapReduce(new HashSet<Integer>(), (lhs, rhs) -> {
                    HashSet<Integer> tmp = new HashSet<>();
                    tmp.addAll(lhs);
                    tmp.addAll(rhs.toSet());
                    return tmp;
                }));
    }

    @Test
    public void testFlatMap() {
        MySet<Integer> tmpA = MySet.range(0, 5);
        MySet<Integer> tmpB = MySet.range(5, 10);
        MySet<Integer> tmpC = MySet.range(10, 15);
        MySet<MySet<Integer>> tmpD = MySet.of(tmpA, tmpB, tmpC);
        assertEquals(MySet.range(0, 15).toSet(),
                tmpD.flatMap(Function.identity()).toSet());
    }

    @Test
    public void testAny() {
        assertTrue(MySet.of(1, 3, 5, 6).any(i -> i % 2 == 0));
        assertFalse(MySet.of(1, 3, 5, 7).any(i -> i % 2 == 0));
    }

    @Test
    public void testAll() {
        assertTrue(MySet.of(1, 3, 5, 7).all(i -> i % 2 != 0));
        assertFalse(MySet.of(2, 4, 5, 6).all(i -> i % 2 == 0));
    }

    @Test
    public void testFilter() {
        assertEquals(Set.of(1, 3, 5, 7, 9), MySet.range(0, 10).filter(i -> i % 2 == 1).toSet());
        assertEquals(Set.of(0, 2, 4, 6, 8), MySet.range(0, 10).filter(i -> i % 2 == 0).toSet());
    }

    @Test
    public void testContains() {
        String a = "abc";
        StringBuilder sb = new StringBuilder();
        sb.append('a');
        sb.append('b');
        sb.append('c');
        assertTrue(MySet.of(a).contains(sb.toString()));
        assertTrue(MySet.of(0, 1, 2, 3, 4).contains(2));
        assertFalse(MySet.of(0, 1, 2, 3, 4).contains(5));
    }

    @Test
    public void testSize() {
        assertSame(8, MySet.range(0, 8).size());
    }

    @Test
    public void testEquals() {
        assertEquals(MySet.of(0, 1, 2, 3, 4), MySet.range(0, 5));
        assertEquals(MySet.of(0, 1), MySet.of(0, 1, 0, 1, 0, 1, 0, 1));
        assertEquals(MySet.of(0, 1, 2), MySet.of(2, 0, 1));
        assertFalse(MySet.of(0, 1, 2, 3).equals(MySet.of(4, 5, 6)));
    }

    @Test
    public void testHashCode() {
        System.out.println("a" + MySet.of(0, 1, 2, 3, 4).hashCode());
        System.out.println("b" + MySet.range(0, 5).hashCode());
        assertEquals(MySet.of(0, 1, 2, 3, 4).hashCode(), MySet.range(0, 5).hashCode());

        assertEquals(MySet.of(0, 1).hashCode(), MySet.of(0, 1, 0, 1, 0, 1, 0, 1).hashCode());
        assertEquals(MySet.of(0, 1, 2).hashCode(), MySet.of(2, 0, 1).hashCode());
        assertTrue(MySet.of(0, 1, 2, 3).hashCode() != MySet.of(4, 5, 6).hashCode());
    }

    @Test
    public void testToString() {
        MySet<Character> tmp = MySet.of('A', 'B', 'C');
        Set<String> expected = Set.of("{A, B, C}", "{A, C, B}", "{B, A, C}", "{B, C, A}", "{C, A, B}", "{C, B, A}");
        assertTrue(expected.contains(tmp.toString()));
    }
}
