package ch.hepia.exos;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.ToDoubleFunction;
import java.util.stream.Collectors;
import java.util.stream.DoubleStream;

import ch.hepia.exos.FiscalDB.Person;

public class Corrections {
    // Tools
    private static double computeMedian(FiscalDB f, ToDoubleFunction<Person> converter) {
        return computeMedian(f.size(), f.parallelStream().mapToDouble(converter));
    }

    private static double computeMedian(Collection<Person> ps, ToDoubleFunction<Person> converter) {
        return computeMedian(ps.size(), ps.parallelStream().mapToDouble(converter));
    }

    private static double computeMedian(int size, DoubleStream s) {
        int mid = size / 2;
        if (size % 2 == 0) {
            return s.sorted()
                    .skip(mid - 1)
                    .limit(2)
                    .average()
                    .getAsDouble();
        } else {
            return s
                    .sorted()
                    .skip(mid)
                    .findFirst()
                    .getAsDouble();
        }
    }

    private static double computeWealth(Person p) {
        return p.getIncomes() + 0.1 * p.getAssets();
    }

    private static double computeTaxes(Person p, double medianIncomes, double medianAssets) {
        double incomes = p.getIncomes();
        double assets = p.getAssets();
        if (p.hasSpouse()) {
            incomes += p.getSpouseOrElseThrow().getIncomes();
            assets += p.getSpouseOrElseThrow().getAssets();
        }
        double incomesRate;
        if (incomes < 0.5 * medianIncomes) {
            incomesRate = 0;
        } else if (incomes <= medianIncomes) {
            incomesRate = 0.12;
        } else if (incomes <= 5 * medianIncomes) {
            incomesRate = 0.33;
        } else { // incomes > 5*median
            incomesRate = 0.5;
        }

        double assetsRate;
        if (assets < 0) {
            assetsRate = -0.1;
        } else if (assets <= medianAssets) {
            assetsRate = 0.01;
        } else if (assets <= 5 * medianIncomes) {
            assetsRate = 0.03;
        } else { // assets > 5*median
            assetsRate = 0.07;
        }

        double incomesTaxes = incomesRate * incomes;
        double assetsTaxes = Math.max(-0.25 * incomesTaxes, assetsRate * assets);
        double taxes = incomesTaxes + assetsTaxes;
        if (p.hasSpouse()) {
            taxes *= 0.5;
        }
        return taxes;
    }

    private static <K, V extends Comparable<V>> Map.Entry<K, V> getMaxEntryByValue(Map<K, V> map) {
        return map.entrySet()
                .parallelStream()
                .collect(Collectors.maxBy(Map.Entry.comparingByValue())).get();
    }

    // Answers
    public static void question1(FiscalDB f) {
        Double medianIncomes = computeMedian(f, Person::getIncomes);
        Double avgIncomes = f.parallelStream()
                .mapToDouble(Person::getIncomes)
                .average()
                .getAsDouble();
        System.out.println("Réponse question 1 :");
        System.out.println("\tRevenu moyen : " + avgIncomes.toString());
        System.out.println("\tRevenu médian : " + medianIncomes.toString());
    }

    public static void question2(FiscalDB f) {
        Person p = f.parallelStream()
                .collect(Collectors.maxBy((p1, p2) -> Double.compare(p1.getIncomes(), p2.getIncomes())))
                .get();
        System.out.println("Réponse question 2 :");
        System.out.println("\tLa personne la plus riche est : " + p.toString());
    }

    public static void question3(FiscalDB f) {
        Person p = f.parallelStream()
                .collect(Collectors.minBy((p1, p2) -> Double.compare(p1.getAssets(), p2.getAssets())))
                .get();
        System.out.println("Réponse question 3 :");
        System.out.println("\tLa personne la plus endettée est : " + p.toString());
    }

    public static void question4(FiscalDB f) {
        int withDebts = (int) f.parallelStream()
                .map(Person::getAssets)
                .filter(assets -> assets < 0)
                .count();
        System.out.println("Réponse question 4 :");
        System.out.println("\tIl y a " + withDebts + " personnes endettées");
    }

    public static void question5(FiscalDB f) {
        int millionaires = (int) f.parallelStream()
                .map(Person::getIncomes)
                .filter(incomes -> incomes > 1e6)
                .count();
        System.out.println("Réponse question 5 :");
        System.out.println("\tIl y a " + millionaires + " personnes qui gagnent plus d'un million");
    }

    public static void question6(FiscalDB f) {
        int married = (int) f.parallelStream()
                .filter(Person::hasSpouse)
                .count();
        System.out.println("Réponse question 6 :");
        System.out.println("\tIl y a " + married + " mariées");
    }

    public static void question7(FiscalDB f) {
        double bottom80Incomes = f.parallelStream()
                .mapToDouble(Person::getIncomes)
                .sorted()
                .limit((long) 0.8 * f.size())
                .reduce(0.0, Double::sum);
        double top5Incomes = f.parallelStream()
                .mapToDouble(Person::getIncomes)
                .sorted()
                .skip((long) (0.95 * f.size()))
                .sum();
        System.out.println("Réponse question 7 :");
        if (bottom80Incomes < top5Incomes) {
            System.out.println(
                    "\tLa somme des revenus des 80% les plus bas est inférieure à celle des revenus des 5% les plus haut");
        } else {
            System.out.println(
                    "\tLa somme des revenus des 80% les plus bas est supérieure à celle des revenus des 5% les plus haut");
        }
    }

    public static void question8(FiscalDB f) {
        int multiMillionaires = (int) f.parallelStream()
                .map(Person::getAssets)
                .filter(assets -> assets >= 2e6)
                .count();
        System.out.println("Réponse question 8 :");
        System.out.println("\tIl y a " + multiMillionaires + " multi-millionnaires");
    }

    public static void question9(FiscalDB f) {
        Person p = f.parallelStream()
                .filter(p1 -> p1.getAssets() >= 2e6)
                .collect(Collectors.minBy((p1, p2) -> Double.compare(p1.getIncomes(), p2
                        .getIncomes())))
                .get();
        System.out.println("Réponse question 9 :");
        System.out.println("\tLe multi-millionnaire qui gagne le moins est : " + p.toString());
    }

    public static void question10(FiscalDB f) {
        Function<Map<Boolean, List<Person>>, Map<Boolean, Double>> personMapToMedianIncomesMap;
        personMapToMedianIncomesMap = m -> m.entrySet()
                .parallelStream()
                .collect(Collectors.toConcurrentMap(Map.Entry::getKey,
                        e -> computeMedian(e.getValue(), Person::getIncomes)));

        Map<Boolean, Double> map = f.parallelStream()
                .collect(Collectors.collectingAndThen(Collectors.groupingBy(Person::hasSpouse),
                        personMapToMedianIncomesMap));

        Double medianMarried = map.get(true);
        Double medianNotMarried = map.get(false);
        System.out.println("Réponse question 10 :");
        if (medianMarried < medianNotMarried) {
            System.out.println("\tLes gens mariés ont un salaire médian moins élevé.");
        } else {
            System.out.println("\tLes gens mariés ont un salaire médian plus élevé.");
        }
    }

    public static void question11(FiscalDB f) {
        double median = computeMedian(f, Corrections::computeWealth);
        System.out.println("Réponse question 11 :");
        System.out.println("\tLa richesse médiane est de : " + median);

    }

    public static void question12(FiscalDB f) {
        Map.Entry<String, Long> entry = f.parallelStream()
                .collect(Collectors.collectingAndThen(
                        Collectors.groupingBy(
                                Person::getCity,
                                Collectors.counting()),
                        Corrections::getMaxEntryByValue));
        System.out.println("Réponse question 12 :");
        System.out.println(
                "\tLa ville la plus peuplée est " + entry.getKey() + " avec " + entry.getValue() + " habitants");
    }

    public static void question13(FiscalDB f) {
        Map.Entry<Character, Double> entry = f.parallelStream()
                .collect(Collectors.collectingAndThen(
                        Collectors.groupingBy(
                                (Person p) -> p.getFirstName().charAt(0),
                                Collectors.summingDouble(Corrections::computeWealth)),
                        Corrections::getMaxEntryByValue));

        System.out.println("Réponse question 13 :");
        System.out.println(
                "\tLes personnes dont le nom commence par " + entry.getKey() + " sont les plus riches");

    }

    public static void question14(FiscalDB f) {
        Predicate<Person> isSingle = Predicate.not(Person::hasSpouse);
        Map.Entry<String, Long> entry = f.parallelStream()
                .filter(isSingle)
                .collect(Collectors.collectingAndThen(
                        Collectors.groupingBy(
                                Person::getCity,
                                Collectors.counting()),
                        Corrections::getMaxEntryByValue));
        System.out.println("Réponse question 14 :");
        System.out.println("\tLa ville qui contient le plus de célibataires est " + entry.getKey() + " avec "
                + entry.getValue() + " célibataires.");
    }

    public static void question15(FiscalDB f) {
        Double threshold = f.parallelStream()
                .mapToDouble(Corrections::computeWealth)
                .sorted()
                .skip((long) (0.95 * f.size()))
                .findFirst()
                .getAsDouble();
        Map.Entry<String, Long> entry = f.parallelStream()
                .filter(p -> computeWealth(p) > threshold)
                .collect(Collectors.collectingAndThen(
                        Collectors.groupingBy(
                                Person::getCity,
                                Collectors.counting()),
                        Corrections::getMaxEntryByValue));
        System.out.println("Réponse question 15 :");
        System.out.println("\tLa ville qui contient le plus de riches est " + entry.getKey() + " avec "
                + entry.getValue() + " riches.");
    }

    public static void question16(FiscalDB f) {
        Function<Map<String, List<Person>>, List<String>> getCitysSortedByMedianWealth = map -> map.entrySet()
                .parallelStream()
                .map(e -> Map.entry(e.getKey(),
                        computeMedian(e.getValue(), Corrections::computeWealth)))
                .sorted(Map.Entry.comparingByValue())
                .map(Map.Entry::getKey)
                .toList();
        List<String> citys = f.parallelStream()
                .collect(Collectors.collectingAndThen(
                        Collectors.groupingBy(Person::getCity),
                        getCitysSortedByMedianWealth));
        System.out.println("Réponse question 16 :");
        System.out.println("\tLes villes par ordre croissant de la médiane des richesses sont :");
        citys.forEach(city -> System.out.println("\t - " + city));
    }

    public static void question17(FiscalDB f) {
        Double median = computeMedian(f, Corrections::computeWealth);
        Function<Double, Integer> classifier = wealth -> {
            if (wealth < median) {
                return 1;
            } else if (wealth < 3 * median) {
                return 2;
            } else if (wealth < 6 * median) {
                return 3;
            } else if (wealth < 12 * median) {
                return 4;
            } else {
                return 5;
            }
        };
        Map<Integer, Long> counts = f.parallelStream()
                .map(Corrections::computeWealth)
                .collect(Collectors.groupingBy(classifier, Collectors.counting()));
        System.out.println("Réponse question 17 :");
        System.out.println("\tLa nombre de personnes par classe est :");
        counts.forEach((cl, count) -> System.out.println("\t - " + cl + ": " + count));
    }

    public static void question18(FiscalDB f) {
        double medianIncomes = computeMedian(f, Person::getIncomes);
        double medianAssets = computeMedian(f, Person::getAssets);
        ToDoubleFunction<Person> computeSinglePersonTaxes = person -> computeTaxes(person, medianIncomes, medianAssets);
        Map.Entry<String, Double> city = f.parallelStream()
                .collect(Collectors.collectingAndThen(
                        Collectors.groupingBy(
                                Person::getCity,
                                Collectors.summingDouble(computeSinglePersonTaxes)),
                        Corrections::getMaxEntryByValue));
        System.out.println("Réponse question 18 :");
        System.out.println("\tLa ville qui paye le plus d'impôt est : " + city.getKey());
    }

    public static void question19(FiscalDB f) {
        double medianIncomes = computeMedian(f, Person::getIncomes);
        double medianAssets = computeMedian(f, Person::getAssets);
        ToDoubleFunction<Person> computeSinglePersonFairness = person -> {
            double taxes = computeTaxes(person, medianIncomes, medianAssets);
            double wealth = computeWealth(person);
            return taxes / wealth;
        };
        Function<Map<String, Double>, List<String>> getCitysSortedByAverageFairness = map -> map.entrySet()
                .parallelStream()
                .sorted(Map.Entry.comparingByValue())
                .map(Map.Entry::getKey)
                .toList();

        List<String> citys = f.parallelStream()
                .collect(Collectors.collectingAndThen(
                        Collectors.groupingBy(
                                Person::getCity,
                                Collectors.averagingDouble(computeSinglePersonFairness)),
                        getCitysSortedByAverageFairness));

        System.out.println("Réponse question 19 :");
        System.out.println("\tLes villes par ordre croissant de justesse moyenne sont :");
        citys.forEach(city -> System.out.println("\t - " + city));
    }

    public static void question20(FiscalDB f) {
        Double median = computeMedian(f, Corrections::computeWealth);
        Function<Double, Integer> classifierWealthClass = wealth -> {
            if (wealth < median) {
                return 1;
            } else if (wealth < 3 * median) {
                return 2;
            } else if (wealth < 6 * median) {
                return 3;
            } else if (wealth < 12 * median) {
                return 4;
            } else {
                return 5;
            }
        };
        Function<Person, String> classifierDiffWealthClassBetweenSpouses = spouse1 -> {
            int classSpouse1 = classifierWealthClass.apply(computeWealth(spouse1));
            int classSpouse2 = classifierWealthClass.apply(computeWealth(spouse1.getSpouseOrElseThrow()));
            int diffClass = Math.abs(classSpouse2 - classSpouse1);
            if (diffClass >= 2) {
                return "Plus d'une classe";
            } else if (diffClass > 0) {
                return "Exactement une classe";
            } else {
                return "De la même classe";
            }
        };
        // The trick here is that we count twice each marriage (once by spouse)
        // Since a ratio is a fraction, it's invariant to multiplication of both terms
        // by the same coefficient (here 2).
        long twiceNumberOfMarriages = f.parallelStream().filter(Person::hasSpouse).collect(Collectors.counting());
        Map<String, Double> percentages = f.parallelStream()
                .filter(Person::hasSpouse)
                .collect(Collectors.collectingAndThen(
                        Collectors.groupingBy(classifierDiffWealthClassBetweenSpouses, Collectors.counting()),
                        m -> m.entrySet().parallelStream()
                                .collect(Collectors.toConcurrentMap(Map.Entry::getKey,
                                        e -> ((double) e.getValue()) / twiceNumberOfMarriages))));
        System.out.println("Réponse question 20 :");
        System.out.println("\tLa répartition des marriages en fonction de la classe de richesse est :");
        percentages
                .forEach((diffClass, percent) -> System.out
                        .println("\t - " + diffClass + " : " + Math.round(1e4 * percent) / 1e2 + "%"));
    }
}
