package ch.hepia.exos;

import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Collection;
import java.util.HashSet;
import java.util.Objects;
import java.util.function.*;

/**
 * A Set implementation, with higher-order methods.
 * 
 * @param <T> The type of stored elements.
 * @see java.util.Set
 */
public class MySet<T> {
    /**
     * The underlying storage.
     */
    private Set<T> storage;

    /**
     * MySet factory.
     * Returns a MySet with the objects passed as paramaters.
     * 
     * @param <T>    The type of the elements to store.
     * @param values The values in the collection.
     * @return The generated collection.
     * @see java.util.Set#of(Object...)
     */
    @SafeVarargs
    public static <T> MySet<T> of(T... values) {
        MySet<T> tmp = new MySet<>();
        for (T value : values) {
            tmp.add(value);
        }
        return tmp;
    }

    /**
     * MySet factory.
     * Returns a MySet of Integers in the range [startInclusive;endExclusive[.
     * 
     * @param startInclusive The lower included bound.
     * @param endExclusive   The upper excluded bound.
     * @return The generated collection.
     */
    public static MySet<Integer> range(int startInclusive, int endExclusive) {
        MySet<Integer> tmp = new MySet<>();
        for (; startInclusive < endExclusive; startInclusive++)
            tmp.add(startInclusive);
        return tmp;
    }

    /**
     * Creates an empty MySet.
     */
    public MySet() {
        storage = new HashSet<T>();
    }

    /**
     * Creates a MySet of T with elements from a given {@link java.util.Collection}
     * of T.
     * 
     * @param elements The {@link java.util.Collection} with the element we want in
     *                 our MySet.
     * @see java.util.Collection
     */
    public MySet(Collection<T> elements) {
        this();
        storage.addAll(elements);
    }

    /**
     * Inserts an element in the MySet.
     * 
     * @param element The element to insert.
     * @return true if the MySet changed, false otherwise (when the element is
     *         already present).
     */
    public boolean add(T element) {
        return storage.add(element);
    }

    /**
     * Inserts all the elements of another MySet.
     * 
     * @param other The MySet with the elements to insert.
     * @return true if the MySet changed (see {@link add}).
     */
    public boolean addAll(MySet<T> other) {
        return storage.addAll(other.toSet());
    }

    /**
     * Empties the MySet.
     */
    public void clear() {
        storage.clear();
    }

    /**
     * Creates a new {@link java.util.Set} with the elements from the MySet.
     * 
     * @return The created Set.
     */
    public Set<T> toSet() {
        return new HashSet<>(storage);
    }

    /**
     * The hashCode is computed by reducing the MySet using
     * {@link java.util.Objects#hash(Object...)} on each element of the MySet.
     * Since the {@link java.util.Objects#hash(Object...)} function is not
     * associative and not commutative, we have to use and intermediate sorted
     * collection.
     */
    @Override
    public int hashCode() {
        SortedSet<T> sortedElements = new TreeSet<>((x, y) -> x.hashCode() - y.hashCode());
        sortedElements.addAll(toSet());

        int hash = size();
        for (Object element : sortedElements) {
            hash = Objects.hash(hash, element);
        }
        return hash;
    }

    /*
     * Les fonctions à remplir commencent ici.
     */
    // Fonctions d'ordre supérieur

    /**
     * Applies a {@link java.util.function.Consumer} on each element in the MySet.
     * 
     * @param action The action to perform, takes an input of type T.
     */
    public void forEach(Consumer<T> action) {
    }

    /**
     * Applies a {@link java.util.function.Function} on each element in the MySet.
     * 
     * @param <U>    The return type of the mapper.
     * @param mapper A {@link java.util.function.Function} that takes an input of
     *               type T and returns an output of type U.
     * @return The MySet containing the results of the element-wise operation.
     */
    public <U> MySet<U> map(Function<T, U> mapper) {
        return new MySet<>();
    }

    /**
     * Reduces a MySet by applying a {@link java.util.function.BinaryOperator}
     * called the accumulator on each
     * element.
     * 
     * @param identity    The neutral element of the accumulator operator.
     * @param accumulator The {@link java.util.function.BinaryOperator} takes two
     *                    inputs of type T and returns
     *                    an output of type U.
     * @return The last result of the accumulator.
     */
    public T reduce(T identity, BinaryOperator<T> accumulator) {
        return identity;
    }

    /**
     * Equivalent to a map followed by a reduce.
     * Reduces a MySet by applying a {@link java.util.function.BiFunction} called
     * the mapperAccumulator
     * on each element.
     * 
     * @param <U>               The return type of the mapperAccumulator.
     * @param identity          The neutral element of the mapperAccumulator
     *                          function.
     * @param mapperAccumulator The {@link java.util.function.BiFunction} takes two
     *                          inputs. The first one
     *                          of type T and the second one of type U. Returns an
     *                          output of type U.
     * @return The last result of the accumulator.
     */
    public <U> U mapReduce(U identity, BiFunction<U, T, U> mapperAccumulator) {
        return identity;
    }

    /**
     * Applies a {@link java.util.function.Function} that generates a MySet on each
     * element in the MySet and concatenates the results.
     * 
     * @param <U>    The type of data stored in the MyMap returned by the mapper.
     * @param mapper A {@link java.util.function.Function} that takes an input of
     *               type T and returns a MySet with elements of type U as
     *               output.
     * @return The concatenation of all the MySets returned by the mapper.
     */
    public <U> MySet<U> flatMap(Function<T, MySet<U>> mapper) {
        return new MySet<>();
    }

    /**
     * Checks that a {@link java.util.function.Predicate} tests true for at least
     * one element of the MySet.
     * 
     * @param predicate The {@link java.util.function.Predicate} to test each
     *                  element.
     * @return true if at least one element tests true.
     */
    public boolean any(Predicate<T> predicate) {
        return false;
    }

    /**
     * Checks that a {@link java.util.function.Predicate} tests true for all
     * elements of the MySet.
     * 
     * @param predicate The {@link java.util.function.Predicate} to test each
     *                  element.
     * @return true if all elements test true.
     */
    public boolean all(Predicate<T> predicate) {
        return false;

    }

    /**
     * Creates a MySet with all the elements that test true with a given
     * {@link java.util.function.Predicate}.
     * 
     * @param predicate The {@link java.util.function.Predicate} to test each
     *                  element.
     * @return the new MySet with the elements that tested true.
     */
    public MySet<T> filter(Predicate<T> predicate) {
        return new MySet<>();
    }

    // Fonctions utilitaires
    /**
     * Checks if the MySet contains a given element.
     * 
     * @param element The element to look for in the MySet.
     * @return true if any element of the MySet is equal to the given parameter.
     */
    public boolean contains(T element) {
        return false;
    }

    /**
     * The size of the MySet.
     * 
     * @return the number of elements in the MySet.
     */
    public int size() {
        return 0;
    }

    /**
     * Two MySets are equal if they contain the same elements.
     */
    @Override
    public boolean equals(Object obj) {
        return false;
    }

    /**
     * Returns the content of the MySet in the format "{e1.toString(),
     * e2.toString(), ..., en.toString()}".
     */
    @Override
    public String toString() {
        return "";
    }
}
