package ch.hepia.exos;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class FiscalDB {
    public class Person {
        private long id;
        private String firstName;
        private String lastName;
        private Optional<Person> spouse;
        private double incomes;
        private double assets;
        private String city;

        private Person(long id, String first, String last, Optional<Person> spouse, double incomes, double assets,
                String city) {
            this.id = id;
            this.firstName = first;
            this.lastName = last;
            this.spouse = spouse;
            this.incomes = incomes;
            this.assets = assets;
            this.city = city;

        }

        private Person(long id, String first, String last, double incomes, double assets,
                String city) {
            this(id, first, last, Optional.empty(), incomes, assets, city);
        }

        private void setSpouse(Person spouse) {
            this.spouse = Optional.of(spouse);
        }

        public long getId() {
            return id;
        }

        public String getFirstName() {
            return firstName;
        }

        public String getLastName() {
            return lastName;
        }

        public Optional<Person> getSpouse() {
            return spouse;
        }

        public boolean hasSpouse() {
            return spouse.isPresent();
        }

        public Person getSpouseOrElseThrow() throws RuntimeException {
            return spouse.orElseThrow(() -> new RuntimeException("This person doesn't have a spouse"));
        }

        public double getIncomes() {
            return incomes;
        }

        public double getAssets() {
            return assets;
        }

        public String getCity() {
            return city;
        }

        @Override
        public int hashCode() {
            return Objects.hash(id);
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj)
                return true;
            if (obj == null || obj.getClass() != this.getClass())
                return false;

            Person otherPerson = (Person) obj;
            return this.id == otherPerson.id &&
                    firstName.equals(otherPerson.firstName) &&
                    lastName.equals(otherPerson.lastName) &&
                    spouse.equals(otherPerson.spouse) &&
                    incomes == otherPerson.incomes &&
                    assets == otherPerson.assets &&
                    city.equals(otherPerson.city);
        }

        @Override
        public String toString() {
            StringBuilder builder = new StringBuilder();
            builder.append("Person (");
            builder.append(firstName + " " + lastName + ", ");
            builder.append("lives in : " + city);
            builder.append(")");
            return builder.toString();
        }
    }

    private Set<Person> db;

    public FiscalDB(String csvFile) throws IOException, IllegalArgumentException {
        try (Stream<String> stream = Files.lines(Paths.get(csvFile)).parallel()) {
            Map<Long, Map.Entry<Long, Person>> table = stream.map(line -> {
                String fields[] = line.split(";");
                if (fields.length != 7) {
                    throw new IllegalArgumentException("The csv file seems ill-formated.");
                }
                long id = Long.parseLong(fields[0]);
                String firstName = fields[1];
                String lastName = fields[2];
                long spouseId = Long.parseLong(fields[3]);
                spouseId = spouseId < 0 ? id : spouseId;
                double incomes = Double.parseDouble(fields[4]);
                double assets = Double.parseDouble(fields[5]);
                String city = fields[6];
                return Map.entry(id, Map.entry(spouseId, new Person(id, firstName, lastName, incomes, assets, city)));
            }).collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));
            Function<Map.Entry<Long, Map.Entry<Long, Person>>, Person> connectSpouses = entry -> {
                Person p = entry.getValue().getValue();
                long spouseId = entry.getValue().getKey();
                if (entry.getKey() != spouseId) {
                    Person spouse = table.get(spouseId).getValue();
                    p.setSpouse(spouse);
                }
                return p;
            };
            this.db = table.entrySet().parallelStream()
                    .map(connectSpouses)
                    .collect(Collectors.toUnmodifiableSet());
        } catch (IllegalArgumentException e) {
            throw e;
        }
    }

    public Stream<Person> stream() {
        return db.stream();
    }

    public Stream<Person> parallelStream() {
        return db.parallelStream();
    }

    public int size() {
        return db.size();
    }
}
